@@ -3,6 +3,11 @@
 #
 # This module is part of GitPython and is released under
 # the BSD License: http://www.opensource.org/licenses/bsd-license.php
+
+from abc import abstractmethod
+from .exc import InvalidGitRepositoryError
+import os.path as osp
+from .compat import is_win
 import contextlib
 from functools import wraps
 import getpass
@@ -16,8 +21,38 @@ import stat
 from sys import maxsize
 import time
 from unittest import SkipTest
+from urllib.parse import urlsplit, urlunsplit
+import warnings
 
-from gitdb.util import (# NOQA @IgnorePep8
+# from git.objects.util import Traversable
+
+# typing ---------------------------------------------------------
+
+from typing import (Any, AnyStr, BinaryIO, Callable, Dict, Generator, IO, Iterator, List,
+                    Optional, Pattern, Sequence, Tuple, TypeVar, Union, cast,
+                    TYPE_CHECKING, overload, )
+
+import pathlib
+
+if TYPE_CHECKING:
+    from git.remote import Remote
+    from git.repo.base import Repo
+    from git.config import GitConfigParser, SectionConstraint
+    from git import Git
+    # from git.objects.base import IndexObject
+
+
+from .types import (Literal, SupportsIndex, Protocol, runtime_checkable,              # because behind py version guards
+                    PathLike, HSH_TD, Total_TD, Files_TD,                            # aliases
+                    Has_id_attribute)
+
+T_IterableObj = TypeVar('T_IterableObj', bound=Union['IterableObj', 'Has_id_attribute'], covariant=True)
+# So IterableList[Head] is subtype of IterableList[IterableObj]
+
+# ---------------------------------------------------------------------
+
+
+from gitdb.util import (  # NOQA @IgnorePep8
     make_sha,
     LockedFD,               # @UnusedImport
     file_contents_ro,       # @UnusedImport
@@ -29,49 +64,50 @@ from gitdb.util import (# NOQA @IgnorePep8
     hex_to_bin,             # @UnusedImport
 )
 
-from git.compat import is_win
-import os.path as osp
-
-from .exc import InvalidGitRepositoryError
-
 
 # NOTE:  Some of the unused imports might be used/imported by others.
 # Handle once test-cases are back up and running.
 # Most of these are unused here, but are for use by git-python modules so these
 # don't see gitdb all the time. Flake of course doesn't like it.
-__all__ = ("stream_copy", "join_path", "to_native_path_windows", "to_native_path_linux",
-           "join_path_native", "Stats", "IndexFileSHA1Writer", "Iterable", "IterableList",
+__all__ = ["stream_copy", "join_path", "to_native_path_linux",
+           "join_path_native", "Stats", "IndexFileSHA1Writer", "IterableObj", "IterableList",
            "BlockingLockFile", "LockFile", 'Actor', 'get_user_id', 'assure_directory_exists',
            'RemoteProgress', 'CallableRemoteProgress', 'rmtree', 'unbare_repo',
-           'HIDE_WINDOWS_KNOWN_ERRORS')
+           'HIDE_WINDOWS_KNOWN_ERRORS']
 
 log = logging.getLogger(__name__)
 
+# types############################################################
+
+
 #: We need an easy way to see if Appveyor TCs start failing,
 #: so the errors marked with this var are considered "acknowledged" ones, awaiting remedy,
 #: till then, we wish to hide them.
 HIDE_WINDOWS_KNOWN_ERRORS = is_win and os.environ.get('HIDE_WINDOWS_KNOWN_ERRORS', True)
 HIDE_WINDOWS_FREEZE_ERRORS = is_win and os.environ.get('HIDE_WINDOWS_FREEZE_ERRORS', True)
 
-#{ Utility Methods
+# { Utility Methods
+
+T = TypeVar('T')
 
 
-def unbare_repo(func):
+def unbare_repo(func: Callable[..., T]) -> Callable[..., T]:
     """Methods with this decorator raise InvalidGitRepositoryError if they
     encounter a bare repository"""
 
     @wraps(func)
-    def wrapper(self, *args, **kwargs):
+    def wrapper(self: 'Remote', *args: Any, **kwargs: Any) -> T:
         if self.repo.bare:
             raise InvalidGitRepositoryError("Method '%s' cannot operate on bare repositories" % func.__name__)
         # END bare method
         return func(self, *args, **kwargs)
     # END wrapper
+
     return wrapper
 
 
 @contextlib.contextmanager
-def cwd(new_dir):
+def cwd(new_dir: PathLike) -> Generator[PathLike, None, None]:
     old_dir = os.getcwd()
     os.chdir(new_dir)
     try:
@@ -80,13 +116,13 @@ def cwd(new_dir):
         os.chdir(old_dir)
 
 
-def rmtree(path):
+def rmtree(path: PathLike) -> None:
     """Remove the given recursively.
 
     :note: we use shutil rmtree but adjust its behaviour to see whether files that
         couldn't be deleted are read-only. Windows will not remove them in that case"""
 
-    def onerror(func, path, exc_info):
+    def onerror(func: Callable, path: PathLike, exc_info: str) -> None:
         # Is the error an access error ?
         os.chmod(path, stat.S_IWUSR)
 
@@ -94,13 +130,13 @@ def rmtree(path):
             func(path)  # Will scream if still not possible to delete.
         except Exception as ex:
             if HIDE_WINDOWS_KNOWN_ERRORS:
-                raise SkipTest("FIXME: fails with: PermissionError\n  {}".format(ex))
+                raise SkipTest("FIXME: fails with: PermissionError\n  {}".format(ex)) from ex
             raise
 
     return shutil.rmtree(path, False, onerror)
 
 
-def rmfile(path):
+def rmfile(path: PathLike) -> None:
     """Ensure file deleted also on *Windows* where read-only files need special treatment."""
     if osp.isfile(path):
         if is_win:
@@ -108,7 +144,7 @@ def rmfile(path):
         os.remove(path)
 
 
-def stream_copy(source, destination, chunk_size=512 * 1024):
+def stream_copy(source: BinaryIO, destination: BinaryIO, chunk_size: int = 512 * 1024) -> int:
     """Copy all data from the source stream into the destination stream in chunks
     of size chunk_size
 
@@ -124,11 +160,12 @@ def stream_copy(source, destination, chunk_size=512 * 1024):
     return br
 
 
-def join_path(a, *p):
+def join_path(a: PathLike, *p: PathLike) -> PathLike:
     """Join path tokens together similar to osp.join, but always use
     '/' instead of possibly '\' on windows."""
-    path = a
+    path = str(a)
     for b in p:
+        b = str(b)
         if not b:
             continue
         if b.startswith('/'):
@@ -142,21 +179,25 @@ def join_path(a, *p):
 
 
 if is_win:
-    def to_native_path_windows(path):
+    def to_native_path_windows(path: PathLike) -> PathLike:
+        path = str(path)
         return path.replace('/', '\\')
 
-    def to_native_path_linux(path):
+    def to_native_path_linux(path: PathLike) -> str:
+        path = str(path)
         return path.replace('\\', '/')
 
+    __all__.append("to_native_path_windows")
     to_native_path = to_native_path_windows
 else:
     # no need for any work on linux
-    def to_native_path_linux(path):
-        return path
+    def to_native_path_linux(path: PathLike) -> str:
+        return str(path)
+
     to_native_path = to_native_path_linux
 
 
-def join_path_native(a, *p):
+def join_path_native(a: PathLike, *p: PathLike) -> PathLike:
     """
     As join path, but makes sure an OS native path is returned. This is only
         needed to play it safe on my dear windows and to assure nice paths that only
@@ -164,7 +205,7 @@ def join_path_native(a, *p):
     return to_native_path(join_path(a, *p))
 
 
-def assure_directory_exists(path, is_file=False):
+def assure_directory_exists(path: PathLike, is_file: bool = False) -> bool:
     """Assure that the directory pointed to by path exists.
 
     :param is_file: If True, path is assumed to be a file and handled correctly.
@@ -174,23 +215,23 @@ def assure_directory_exists(path, is_file=False):
         path = osp.dirname(path)
     # END handle file
     if not osp.isdir(path):
-        os.makedirs(path)
+        os.makedirs(path, exist_ok=True)
         return True
     return False
 
 
-def _get_exe_extensions():
+def _get_exe_extensions() -> Sequence[str]:
     PATHEXT = os.environ.get('PATHEXT', None)
-    return tuple(p.upper() for p in PATHEXT.split(os.pathsep)) \
-        if PATHEXT \
-        else (('.BAT', 'COM', '.EXE') if is_win else ())
+    return tuple(p.upper() for p in PATHEXT.split(os.pathsep)) if PATHEXT \
+        else ('.BAT', 'COM', '.EXE') if is_win \
+        else ('')
 
 
-def py_where(program, path=None):
+def py_where(program: str, path: Optional[PathLike] = None) -> List[str]:
     # From: http://stackoverflow.com/a/377028/548792
     winprog_exts = _get_exe_extensions()
 
-    def is_exec(fpath):
+    def is_exec(fpath: str) -> bool:
         return osp.isfile(fpath) and os.access(fpath, os.X_OK) and (
             os.name != 'nt' or not winprog_exts or any(fpath.upper().endswith(ext)
                                                        for ext in winprog_exts))
@@ -198,7 +239,7 @@ def py_where(program, path=None):
     progs = []
     if not path:
         path = os.environ["PATH"]
-    for folder in path.split(os.pathsep):
+    for folder in str(path).split(os.pathsep):
         folder = folder.strip('"')
         if folder:
             exe_path = osp.join(folder, program)
@@ -208,11 +249,11 @@ def py_where(program, path=None):
     return progs
 
 
-def _cygexpath(drive, path):
+def _cygexpath(drive: Optional[str], path: str) -> str:
     if osp.isabs(path) and not drive:
-        ## Invoked from `cygpath()` directly with `D:Apps\123`?
+        # Invoked from `cygpath()` directly with `D:Apps\123`?
         #  It's an error, leave it alone just slashes)
-        p = path
+        p = path  # convert to str if AnyPath given
     else:
         p = path and osp.normpath(osp.expandvars(osp.expanduser(path)))
         if osp.isabs(p):
@@ -223,40 +264,44 @@ def _cygexpath(drive, path):
                 p = cygpath(p)
         elif drive:
             p = '/cygdrive/%s/%s' % (drive.lower(), p)
-
-    return p.replace('\\', '/')
+    p_str = str(p)  # ensure it is a str and not AnyPath
+    return p_str.replace('\\', '/')
 
 
-_cygpath_parsers = (
-    ## See: https://msdn.microsoft.com/en-us/library/windows/desktop/aa365247(v=vs.85).aspx
-    ## and: https://www.cygwin.com/cygwin-ug-net/using.html#unc-paths
+_cygpath_parsers: Tuple[Tuple[Pattern[str], Callable, bool], ...] = (
+    # See: https://msdn.microsoft.com/en-us/library/windows/desktop/aa365247(v=vs.85).aspx
+    # and: https://www.cygwin.com/cygwin-ug-net/using.html#unc-paths
     (re.compile(r"\\\\\?\\UNC\\([^\\]+)\\([^\\]+)(?:\\(.*))?"),
      (lambda server, share, rest_path: '//%s/%s/%s' % (server, share, rest_path.replace('\\', '/'))),
      False
      ),
 
     (re.compile(r"\\\\\?\\(\w):[/\\](.*)"),
-     _cygexpath,
+     (_cygexpath),
      False
      ),
 
     (re.compile(r"(\w):[/\\](.*)"),
-     _cygexpath,
+     (_cygexpath),
      False
      ),
 
     (re.compile(r"file:(.*)", re.I),
      (lambda rest_path: rest_path),
-     True),
+     True
+     ),
 
     (re.compile(r"(\w{2,}:.*)"),  # remote URL, do nothing
      (lambda url: url),
-     False),
+     False
+     ),
 )
 
 
-def cygpath(path):
+def cygpath(path: str) -> str:
     """Use :meth:`git.cmd.Git.polish_url()` instead, that works on any environment."""
+    path = str(path)  # ensure is str and not AnyPath.
+    # Fix to use Paths when 3.5 dropped. or to be just str if only for urls?
     if not path.startswith(('/cygdrive', '//')):
         for regex, parser, recurse in _cygpath_parsers:
             match = regex.match(path)
@@ -274,7 +319,8 @@ def cygpath(path):
 _decygpath_regex = re.compile(r"/cygdrive/(\w)(/.*)?")
 
 
-def decygpath(path):
+def decygpath(path: PathLike) -> str:
+    path = str(path)
     m = _decygpath_regex.match(path)
     if m:
         drive, rest_path = m.groups()
@@ -285,25 +331,37 @@ def decygpath(path):
 
 #: Store boolean flags denoting if a specific Git executable
 #: is from a Cygwin installation (since `cache_lru()` unsupported on PY2).
-_is_cygwin_cache = {}
+_is_cygwin_cache: Dict[str, Optional[bool]] = {}
 
 
-def is_cygwin_git(git_executable):
+@overload
+def is_cygwin_git(git_executable: None) -> Literal[False]:
+    ...
+
+
+@overload
+def is_cygwin_git(git_executable: PathLike) -> bool:
+    ...
+
+
+def is_cygwin_git(git_executable: Union[None, PathLike]) -> bool:
     if not is_win:
         return False
 
-    #from subprocess import check_output
+    if git_executable is None:
+        return False
 
-    is_cygwin = _is_cygwin_cache.get(git_executable)
+    git_executable = str(git_executable)
+    is_cygwin = _is_cygwin_cache.get(git_executable)  # type: Optional[bool]
     if is_cygwin is None:
         is_cygwin = False
         try:
             git_dir = osp.dirname(git_executable)
             if not git_dir:
                 res = py_where(git_executable)
-                git_dir = osp.dirname(res[0]) if res else None
+                git_dir = osp.dirname(res[0]) if res else ""
 
-            ## Just a name given, not a real path.
+            # Just a name given, not a real path.
             uname_cmd = osp.join(git_dir, 'uname')
             process = subprocess.Popen([uname_cmd], stdout=subprocess.PIPE,
                                        universal_newlines=True)
@@ -317,29 +375,70 @@ def is_cygwin_git(git_executable):
     return is_cygwin
 
 
-def get_user_id():
+def get_user_id() -> str:
     """:return: string identifying the currently active system user as name@node"""
     return "%s@%s" % (getpass.getuser(), platform.node())
 
 
-def finalize_process(proc, **kwargs):
+def finalize_process(proc: Union[subprocess.Popen, 'Git.AutoInterrupt'], **kwargs: Any) -> None:
     """Wait for the process (clone, fetch, pull or push) and handle its errors accordingly"""
-    ## TODO: No close proc-streams??
+    # TODO: No close proc-streams??
     proc.wait(**kwargs)
 
 
-def expand_path(p, expand_vars=True):
+@overload
+def expand_path(p: None, expand_vars: bool = ...) -> None:
+    ...
+
+
+@overload
+def expand_path(p: PathLike, expand_vars: bool = ...) -> str:
+    # improve these overloads when 3.5 dropped
+    ...
+
+
+def expand_path(p: Union[None, PathLike], expand_vars: bool = True) -> Optional[PathLike]:
+    if isinstance(p, pathlib.Path):
+        return p.resolve()
     try:
-        p = osp.expanduser(p)
+        p = osp.expanduser(p)  # type: ignore
         if expand_vars:
-            p = osp.expandvars(p)
-        return osp.normpath(osp.abspath(p))
+            p = osp.expandvars(p)    # type: ignore
+        return osp.normpath(osp.abspath(p))  # type: ignore
     except Exception:
         return None
 
-#} END utilities
 
-#{ Classes
+def remove_password_if_present(cmdline: Sequence[str]) -> List[str]:
+    """
+    Parse any command line argument and if on of the element is an URL with a
+    password, replace it by stars (in-place).
+
+    If nothing found just returns the command line as-is.
+
+    This should be used for every log line that print a command line.
+    """
+    new_cmdline = []
+    for index, to_parse in enumerate(cmdline):
+        new_cmdline.append(to_parse)
+        try:
+            url = urlsplit(to_parse)
+            # Remove password from the URL if present
+            if url.password is None:
+                continue
+
+            edited_url = url._replace(
+                netloc=url.netloc.replace(url.password, "*****"))
+            new_cmdline[index] = urlunsplit(edited_url)
+        except ValueError:
+            # This is not a valid URL
+            continue
+    return new_cmdline
+
+
+# } END utilities
+
+# { Classes
 
 
 class RemoteProgress(object):
@@ -347,7 +446,7 @@ class RemoteProgress(object):
     Handler providing an interface to parse progress information emitted by git-push
     and git-fetch and to dispatch callbacks allowing subclasses to react to the progress.
     """
-    _num_op_codes = 9
+    _num_op_codes: int = 9
     BEGIN, END, COUNTING, COMPRESSING, WRITING, RECEIVING, RESOLVING, FINDING_SOURCES, CHECKING_OUT = \
         [1 << x for x in range(_num_op_codes)]
     STAGE_MASK = BEGIN | END
@@ -363,51 +462,56 @@ class RemoteProgress(object):
     re_op_absolute = re.compile(r"(remote: )?([\w\s]+):\s+()(\d+)()(.*)")
     re_op_relative = re.compile(r"(remote: )?([\w\s]+):\s+(\d+)% \((\d+)/(\d+)\)(.*)")
 
-    def __init__(self):
-        self._seen_ops = []
-        self._cur_line = None
-        self.error_lines = []
-        self.other_lines = []
+    def __init__(self) -> None:
+        self._seen_ops: List[int] = []
+        self._cur_line: Optional[str] = None
+        self.error_lines: List[str] = []
+        self.other_lines: List[str] = []
 
-    def _parse_progress_line(self, line):
+    def _parse_progress_line(self, line: AnyStr) -> None:
         """Parse progress information from the given line as retrieved by git-push
         or git-fetch.
 
         - Lines that do not contain progress info are stored in :attr:`other_lines`.
         - Lines that seem to contain an error (i.e. start with error: or fatal:) are stored
-        in :attr:`error_lines`."""
+            in :attr:`error_lines`."""
         # handle
         # Counting objects: 4, done.
         # Compressing objects:  50% (1/2)
         # Compressing objects: 100% (2/2)
         # Compressing objects: 100% (2/2), done.
-        self._cur_line = line = line.decode('utf-8') if isinstance(line, bytes) else line
-        if self.error_lines or self._cur_line.startswith(('error:', 'fatal:')):
+        if isinstance(line, bytes):   # mypy argues about ternary assignment
+            line_str = line.decode('utf-8')
+        else:
+            line_str = line
+        self._cur_line = line_str
+
+        if self._cur_line.startswith(('error:', 'fatal:')):
             self.error_lines.append(self._cur_line)
             return
 
         # find escape characters and cut them away - regex will not work with
         # them as they are non-ascii. As git might expect a tty, it will send them
         last_valid_index = None
-        for i, c in enumerate(reversed(line)):
+        for i, c in enumerate(reversed(line_str)):
             if ord(c) < 32:
                 # its a slice index
                 last_valid_index = -i - 1
             # END character was non-ascii
         # END for each character in line
         if last_valid_index is not None:
-            line = line[:last_valid_index]
+            line_str = line_str[:last_valid_index]
         # END cut away invalid part
-        line = line.rstrip()
+        line_str = line_str.rstrip()
 
         cur_count, max_count = None, None
-        match = self.re_op_relative.match(line)
+        match = self.re_op_relative.match(line_str)
         if match is None:
-            match = self.re_op_absolute.match(line)
+            match = self.re_op_absolute.match(line_str)
 
         if not match:
-            self.line_dropped(line)
-            self.other_lines.append(line)
+            self.line_dropped(line_str)
+            self.other_lines.append(line_str)
             return
         # END could not get match
 
@@ -436,10 +540,10 @@ class RemoteProgress(object):
             # This can't really be prevented, so we drop the line verbosely
             # to make sure we get informed in case the process spits out new
             # commands at some point.
-            self.line_dropped(line)
+            self.line_dropped(line_str)
             # Note: Don't add this line to the other lines, as we have to silently
             # drop it
-            return
+            return None
         # END handle op code
 
         # figure out stage
@@ -464,21 +568,22 @@ class RemoteProgress(object):
                     max_count and float(max_count),
                     message)
 
-    def new_message_handler(self):
+    def new_message_handler(self) -> Callable[[str], None]:
         """
         :return:
             a progress handler suitable for handle_process_output(), passing lines on to this Progress
             handler in a suitable format"""
-        def handler(line):
+        def handler(line: AnyStr) -> None:
             return self._parse_progress_line(line.rstrip())
         # end
         return handler
 
-    def line_dropped(self, line):
+    def line_dropped(self, line: str) -> None:
         """Called whenever a line could not be understood and was therefore dropped."""
         pass
 
-    def update(self, op_code, cur_count, max_count=None, message=''):
+    def update(self, op_code: int, cur_count: Union[str, float], max_count: Union[str, float, None] = None,
+               message: str = '',) -> None:
         """Called whenever the progress changes
 
         :param op_code:
@@ -509,11 +614,11 @@ class CallableRemoteProgress(RemoteProgress):
     """An implementation forwarding updates to any callable"""
     __slots__ = ('_callable')
 
-    def __init__(self, fn):
+    def __init__(self, fn: Callable) -> None:
         self._callable = fn
         super(CallableRemoteProgress, self).__init__()
 
-    def update(self, *args, **kwargs):
+    def update(self, *args: Any, **kwargs: Any) -> None:
         self._callable(*args, **kwargs)
 
 
@@ -538,27 +643,27 @@ class Actor(object):
 
     __slots__ = ('name', 'email')
 
-    def __init__(self, name, email):
+    def __init__(self, name: Optional[str], email: Optional[str]) -> None:
         self.name = name
         self.email = email
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any) -> bool:
         return self.name == other.name and self.email == other.email
 
-    def __ne__(self, other):
+    def __ne__(self, other: Any) -> bool:
         return not (self == other)
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return hash((self.name, self.email))
 
-    def __str__(self):
-        return self.name
+    def __str__(self) -> str:
+        return self.name if self.name else ""
 
-    def __repr__(self):
-        return u'<git.Actor "%s <%s>">' % (self.name, self.email)
+    def __repr__(self) -> str:
+        return '<git.Actor "%s <%s>">' % (self.name, self.email)
 
     @classmethod
-    def _from_string(cls, string):
+    def _from_string(cls, string: str) -> 'Actor':
         """Create an Actor from a string.
         :param string: is the string, which is expected to be in regular git format
 
@@ -579,10 +684,19 @@ class Actor(object):
         # END handle name/email matching
 
     @classmethod
-    def _main_actor(cls, env_name, env_email, config_reader=None):
+    def _main_actor(cls, env_name: str, env_email: str,
+                    config_reader: Union[None, 'GitConfigParser', 'SectionConstraint'] = None) -> 'Actor':
         actor = Actor('', '')
-        default_email = get_user_id()
-        default_name = default_email.split('@')[0]
+        user_id = None  # We use this to avoid multiple calls to getpass.getuser()
+
+        def default_email() -> str:
+            nonlocal user_id
+            if not user_id:
+                user_id = get_user_id()
+            return user_id
+
+        def default_name() -> str:
+            return default_email().split('@')[0]
 
         for attr, evar, cvar, default in (('name', env_name, cls.conf_name, default_name),
                                           ('email', env_email, cls.conf_email, default_email)):
@@ -591,16 +705,20 @@ class Actor(object):
                 setattr(actor, attr, val)
             except KeyError:
                 if config_reader is not None:
-                    setattr(actor, attr, config_reader.get_value('user', cvar, default))
+                    try:
+                        val = config_reader.get('user', cvar)
+                    except Exception:
+                        val = default()
+                    setattr(actor, attr, val)
                 # END config-reader handling
                 if not getattr(actor, attr):
-                    setattr(actor, attr, default)
+                    setattr(actor, attr, default())
             # END handle name
         # END for each item to retrieve
         return actor
 
     @classmethod
-    def committer(cls, config_reader=None):
+    def committer(cls, config_reader: Union[None, 'GitConfigParser', 'SectionConstraint'] = None) -> 'Actor':
         """
         :return: Actor instance corresponding to the configured committer. It behaves
             similar to the git implementation, such that the environment will override
@@ -611,7 +729,7 @@ class Actor(object):
         return cls._main_actor(cls.env_committer_name, cls.env_committer_email, config_reader)
 
     @classmethod
-    def author(cls, config_reader=None):
+    def author(cls, config_reader: Union[None, 'GitConfigParser', 'SectionConstraint'] = None) -> 'Actor':
         """Same as committer(), but defines the main author. It may be specified in the environment,
         but defaults to the committer"""
         return cls._main_actor(cls.env_author_name, cls.env_author_email, config_reader)
@@ -645,16 +763,22 @@ class Stats(object):
      files = number of changed files as int"""
     __slots__ = ("total", "files")
 
-    def __init__(self, total, files):
+    def __init__(self, total: Total_TD, files: Dict[PathLike, Files_TD]):
         self.total = total
         self.files = files
 
     @classmethod
-    def _list_from_string(cls, repo, text):
+    def _list_from_string(cls, repo: 'Repo', text: str) -> 'Stats':
         """Create a Stat object from output retrieved by git-diff.
 
         :return: git.Stat"""
-        hsh = {'total': {'insertions': 0, 'deletions': 0, 'lines': 0, 'files': 0}, 'files': {}}
+
+        hsh: HSH_TD = {'total': {'insertions': 0,
+                                 'deletions': 0,
+                                 'lines': 0,
+                                 'files': 0},
+                       'files': {}
+                       }
         for line in text.splitlines():
             (raw_insertions, raw_deletions, filename) = line.split("\t")
             insertions = raw_insertions != '-' and int(raw_insertions) or 0
@@ -663,9 +787,10 @@ class Stats(object):
             hsh['total']['deletions'] += deletions
             hsh['total']['lines'] += insertions + deletions
             hsh['total']['files'] += 1
-            hsh['files'][filename.strip()] = {'insertions': insertions,
-                                              'deletions': deletions,
-                                              'lines': insertions + deletions}
+            files_dict: Files_TD = {'insertions': insertions,
+                                    'deletions': deletions,
+                                    'lines': insertions + deletions}
+            hsh['files'][filename.strip()] = files_dict
         return Stats(hsh['total'], hsh['files'])
 
 
@@ -680,25 +805,25 @@ class IndexFileSHA1Writer(object):
     :note: Based on the dulwich project"""
     __slots__ = ("f", "sha1")
 
-    def __init__(self, f):
+    def __init__(self, f: IO) -> None:
         self.f = f
         self.sha1 = make_sha(b"")
 
-    def write(self, data):
+    def write(self, data: AnyStr) -> int:
         self.sha1.update(data)
         return self.f.write(data)
 
-    def write_sha(self):
+    def write_sha(self) -> bytes:
         sha = self.sha1.digest()
         self.f.write(sha)
         return sha
 
-    def close(self):
+    def close(self) -> bytes:
         sha = self.write_sha()
         self.f.close()
         return sha
 
-    def tell(self):
+    def tell(self) -> int:
         return self.f.tell()
 
 
@@ -712,23 +837,23 @@ class LockFile(object):
     Locks will automatically be released on destruction"""
     __slots__ = ("_file_path", "_owns_lock")
 
-    def __init__(self, file_path):
+    def __init__(self, file_path: PathLike) -> None:
         self._file_path = file_path
         self._owns_lock = False
 
-    def __del__(self):
+    def __del__(self) -> None:
         self._release_lock()
 
-    def _lock_file_path(self):
+    def _lock_file_path(self) -> str:
         """:return: Path to lockfile"""
         return "%s.lock" % (self._file_path)
 
-    def _has_lock(self):
+    def _has_lock(self) -> bool:
         """:return: True if we have a lock and if the lockfile still exists
         :raise AssertionError: if our lock-file does not exist"""
         return self._owns_lock
 
-    def _obtain_lock_or_raise(self):
+    def _obtain_lock_or_raise(self) -> None:
         """Create a lock file as flag for other instances, mark our instance as lock-holder
 
         :raise IOError: if a lock was already present or a lock file could not be written"""
@@ -746,16 +871,16 @@ class LockFile(object):
             fd = os.open(lock_file, flags, 0)
             os.close(fd)
         except OSError as e:
-            raise IOError(str(e))
+            raise IOError(str(e)) from e
 
         self._owns_lock = True
 
-    def _obtain_lock(self):
+    def _obtain_lock(self) -> None:
         """The default implementation will raise if a lock cannot be obtained.
         Subclasses may override this method to provide a different implementation"""
         return self._obtain_lock_or_raise()
 
-    def _release_lock(self):
+    def _release_lock(self) -> None:
         """Release our lock if we have one"""
         if not self._has_lock():
             return
@@ -780,7 +905,7 @@ class BlockingLockFile(LockFile):
         can never be obtained."""
     __slots__ = ("_check_interval", "_max_block_time")
 
-    def __init__(self, file_path, check_interval_s=0.3, max_block_time_s=maxsize):
+    def __init__(self, file_path: PathLike, check_interval_s: float = 0.3, max_block_time_s: int = maxsize) -> None:
         """Configure the instance
 
         :param check_interval_s:
@@ -792,7 +917,7 @@ class BlockingLockFile(LockFile):
         self._check_interval = check_interval_s
         self._max_block_time = max_block_time_s
 
-    def _obtain_lock(self):
+    def _obtain_lock(self) -> None:
         """This method blocks until it obtained the lock, or raises IOError if
         it ran out of time or if the parent directory was not available anymore.
         If this method returns, you are guaranteed to own the lock"""
@@ -801,19 +926,19 @@ class BlockingLockFile(LockFile):
         while True:
             try:
                 super(BlockingLockFile, self)._obtain_lock()
-            except IOError:
+            except IOError as e:
                 # synity check: if the directory leading to the lockfile is not
                 # readable anymore, raise an exception
                 curtime = time.time()
                 if not osp.isdir(osp.dirname(self._lock_file_path())):
                     msg = "Directory containing the lockfile %r was not readable anymore after waiting %g seconds" % (
                         self._lock_file_path(), curtime - starttime)
-                    raise IOError(msg)
+                    raise IOError(msg) from e
                 # END handle missing directory
 
                 if curtime >= maxtime:
                     msg = "Waited %g seconds for lock at %r" % (maxtime - starttime, self._lock_file_path())
-                    raise IOError(msg)
+                    raise IOError(msg) from e
                 # END abort if we wait too long
                 time.sleep(self._check_interval)
             else:
@@ -821,7 +946,7 @@ class BlockingLockFile(LockFile):
         # END endless loop
 
 
-class IterableList(list):
+class IterableList(List[T_IterableObj]):
 
     """
     List of iterable objects allowing to query an object by id or by named index::
@@ -831,6 +956,9 @@ class IterableList(list):
      heads['master']
      heads[0]
 
+    Iterable parent objects = [Commit, SubModule, Reference, FetchInfo, PushInfo]
+    Iterable via inheritance = [Head, TagReference, RemoteReference]
+    ]
     It requires an id_attribute name to be set which will be queried from its
     contained items to have a means for comparison.
 
@@ -839,14 +967,14 @@ class IterableList(list):
     can be left out."""
     __slots__ = ('_id_attr', '_prefix')
 
-    def __new__(cls, id_attr, prefix=''):
+    def __new__(cls, id_attr: str, prefix: str = '') -> 'IterableList[IterableObj]':
         return super(IterableList, cls).__new__(cls)
 
-    def __init__(self, id_attr, prefix=''):
+    def __init__(self, id_attr: str, prefix: str = '') -> None:
         self._id_attr = id_attr
         self._prefix = prefix
 
-    def __contains__(self, attr):
+    def __contains__(self, attr: object) -> bool:
         # first try identity match for performance
         try:
             rval = list.__contains__(self, attr)
@@ -858,13 +986,13 @@ class IterableList(list):
 
         # otherwise make a full name search
         try:
-            getattr(self, attr)
+            getattr(self, cast(str, attr))  # use cast to silence mypy
             return True
         except (AttributeError, TypeError):
             return False
         # END handle membership
 
-    def __getattr__(self, attr):
+    def __getattr__(self, attr: str) -> T_IterableObj:
         attr = self._prefix + attr
         for item in self:
             if getattr(item, self._id_attr) == attr:
@@ -872,18 +1000,26 @@ class IterableList(list):
         # END for each item
         return list.__getattribute__(self, attr)
 
-    def __getitem__(self, index):
+    def __getitem__(self, index: Union[SupportsIndex, int, slice, str]) -> T_IterableObj:  # type: ignore
+
+        assert isinstance(index, (int, str, slice)), "Index of IterableList should be an int or str"
+
         if isinstance(index, int):
             return list.__getitem__(self, index)
-
-        try:
-            return getattr(self, index)
-        except AttributeError:
-            raise IndexError("No item found with id %r" % (self._prefix + index))
+        elif isinstance(index, slice):
+            raise ValueError("Index should be an int or str")
+        else:
+            try:
+                return getattr(self, index)
+            except AttributeError as e:
+                raise IndexError("No item found with id %r" % (self._prefix + index)) from e
         # END handle getattr
 
-    def __delitem__(self, index):
-        delindex = index
+    def __delitem__(self, index: Union[SupportsIndex, int, slice, str]) -> None:
+
+        assert isinstance(index, (int, str)), "Index of IterableList should be an int or str"
+
+        delindex = cast(int, index)
         if not isinstance(index, int):
             delindex = -1
             name = self._prefix + index
@@ -900,7 +1036,20 @@ class IterableList(list):
         list.__delitem__(self, delindex)
 
 
-class Iterable(object):
+class IterableClassWatcher(type):
+    """ Metaclass that watches """
+    def __init__(cls, name: str, bases: Tuple, clsdict: Dict) -> None:
+        for base in bases:
+            if type(base) == IterableClassWatcher:
+                warnings.warn(f"GitPython Iterable subclassed by {name}. "
+                              "Iterable is deprecated due to naming clash since v3.1.18"
+                              " and will be removed in 3.1.20, "
+                              "Use IterableObj instead \n",
+                              DeprecationWarning,
+                              stacklevel=2)
+
+
+class Iterable(metaclass=IterableClassWatcher):
 
     """Defines an interface for iterable items which is to assure a uniform
     way to retrieve and iterate items within the git repository"""
@@ -908,7 +1057,40 @@ class Iterable(object):
     _id_attribute_ = "attribute that most suitably identifies your instance"
 
     @classmethod
-    def list_items(cls, repo, *args, **kwargs):
+    def list_items(cls, repo: 'Repo', *args: Any, **kwargs: Any) -> Any:
+        """
+        Deprecated, use IterableObj instead.
+        Find all items of this type - subclasses can specify args and kwargs differently.
+        If no args are given, subclasses are obliged to return all items if no additional
+        arguments arg given.
+
+        :note: Favor the iter_items method as it will
+
+        :return:list(Item,...) list of item instances"""
+        out_list: Any = IterableList(cls._id_attribute_)
+        out_list.extend(cls.iter_items(repo, *args, **kwargs))
+        return out_list
+
+    @classmethod
+    def iter_items(cls, repo: 'Repo', *args: Any, **kwargs: Any) -> Any:
+        # return typed to be compatible with subtypes e.g. Remote
+        """For more information about the arguments, see list_items
+        :return:  iterator yielding Items"""
+        raise NotImplementedError("To be implemented by Subclass")
+
+
+@runtime_checkable
+class IterableObj(Protocol):
+    """Defines an interface for iterable items which is to assure a uniform
+    way to retrieve and iterate items within the git repository
+
+    Subclasses = [Submodule, Commit, Reference, PushInfo, FetchInfo, Remote]"""
+
+    __slots__ = ()
+    _id_attribute_: str
+
+    @classmethod
+    def list_items(cls, repo: 'Repo', *args: Any, **kwargs: Any) -> IterableList[T_IterableObj]:
         """
         Find all items of this type - subclasses can specify args and kwargs differently.
         If no args are given, subclasses are obliged to return all items if no additional
@@ -917,19 +1099,22 @@ class Iterable(object):
         :note: Favor the iter_items method as it will
 
         :return:list(Item,...) list of item instances"""
-        out_list = IterableList(cls._id_attribute_)
+        out_list: IterableList = IterableList(cls._id_attribute_)
         out_list.extend(cls.iter_items(repo, *args, **kwargs))
         return out_list
 
     @classmethod
-    def iter_items(cls, repo, *args, **kwargs):
+    @abstractmethod
+    def iter_items(cls, repo: 'Repo', *args: Any, **kwargs: Any
+                   ) -> Iterator[T_IterableObj]:  # Iterator[T_IterableObj]:
+        # return typed to be compatible with subtypes e.g. Remote
         """For more information about the arguments, see list_items
-        :return:  iterator yielding Items"""
+            :return:  iterator yielding Items"""
         raise NotImplementedError("To be implemented by Subclass")
 
-#} END classes
+# } END classes
 
 
 class NullHandler(logging.Handler):
-    def emit(self, record):
+    def emit(self, record: object) -> None:
         pass
